<?php
class Rentsoft
{
    const VERSION = "2.01";

    /**
     * Вызывается на стороне Агента
     * 
     * Возвращает код генерации элемента iframe для размещения в личном кабинете Агента.
     * Через iframe пользователю доступен интерфейс управления услугами Рентсофт
     *
     * @param string $rsUri             GET-параметр самой верхней страницы, открытой в браузере
     * @param string $agRef             Аргумент сохранён для обратной совместимости (передавайте null)
     * @param string $agName            Системное имя Агента
     * @param string $agUuid            ID пользователя в системе Агента
     * @param string $agApi             Полный URL API Агента, который позволяет работать с этим пользователем
     * @param string $agSecret          Секретный ключ, используется для цифровой подписи
     * @param string $devDomainSuffix   Используется для отладки (по умолчанию '')
     * @param int $width                Ширина IFRAME (по умолчанию 100%)
     * @return string                   Код генерации элемента iframe
     */
    public static function getIframe($rsUri, $agRef, $agName, $agUuid, $agApi, $agSecret, $devDomainSuffix = '', $width = null)
    {
        // Since 2010-11-15 $agRef is not used, but it is still in the 
        // interface to keep function prototype backward-compatible.
                    
        // Build and sign response params.
        $rsSignedParams = 
            "ag_uuid=" . urlencode($agUuid) . "&" .     // agent's user ID 
            "ag_api=" . urlencode($agApi) . "&" .       // agent's API for this user
            "ag_timestamp=" . time() . "&" .            // timestamp of this request creation WITH MILLISECONDS
            "ag_rnd=" . mt_rand();                      // a random number to get unique URL
        $rsSignedParams .= "&ag_sign=" . md5($agSecret . $rsSignedParams);

        // Build response URI.        
        $rsResponseUri = '/' . self::VERSION . "/iframe/" . ltrim($rsUri, '/')
            . (false !== strpos($rsUri, '?')? '&' : '?') 
            . $rsSignedParams;

        // Build IFRAME hostname.
        $rsHostname = "{$agName}.ag.rentsoft.ru{$devDomainSuffix}";
        
        // Build IFRAME full URL.
        // Note that width must be set by JS together with height due to IE bug.
        $proto = preg_match('/\.(dev|test|prod)\./si', $rsHostname) ? "http" : "https";
        return self::getStretchingIframe('rentsoft_ag', "{$proto}://{$rsHostname}{$rsResponseUri}", $width);
    }

    /**
     * Вызывается на стороне RentSoft.
     * Исходный код приведён для примера
     * 
     * Разбирает REQUEST_URI и HTTP_HOST, проверяет цифровую подпись iframe и т.д.
     *
     * @param string $httpHost
     * @param string $requestUri
     * @param string $rsSecretCallback  Callback для получения секретного ключа по имени агента
     * @param callback $nonceCallback   Callback для контроля повторной отправки запросов
     * @return array                    Разобранные аргументы
     */
    public static function readArgs($httpHost, $requestUri, $rsSecretCallback, $nonceCallback)
    {
        $agName = preg_replace('/\..*$/s', '', $httpHost);
        $rsSecret = call_user_func($rsSecretCallback, $agName);
        
        $expireDelta = 2 * 24 * 3600; // must be > 2 days, because we know nothing about time zone here
        $args = array();
        parse_str(parse_url($requestUri, PHP_URL_QUERY), $args);
        $prefix = preg_replace('/&ag_sign=.*$/s', '', $requestUri);

        if (md5($rsSecret . $prefix) !== @$args['ag_sign']) {
            throw new Exception("Invalid digital signature!");
        }
        $agTimestamp = @$args['ag_timestamp'];
        $dt = time() - $agTimestamp;
        // Check if the request is knowingly expired.
        if ($dt > $expireDelta) {
            throw new Exception("The request is expired (age is $dt seconds which is more than $expireDelta seconds)");
        }
        // Check nonce (typically all used URIs are stored in DB).
        if (!call_user_func($nonceCallback, $prefix)) {
            throw new Exception("This request was already used: $prefix");
        }
        return array(
            'agName' => $agName,
            'agUuid' => @$args['ag_uuid'],
            'agWsdl' => @$args['ag_wsdl'],
            'agRef'  => @$args['ag_ref'],
            'agTimestamp' => $agTimestamp,
            'agSign' => @$args['ag_sign'],
        );
    }

    /**
     * Вызывается на стороне Агента
     *
     * Возвращает код генерации элемента iframe с поддержкой автоматической подстройки высоты.
     * Может использоваться для размещения витрины Рентсофт внутри iframe'а
     *
     * @param  string $id          Префикс атрибута id элемента iframe и его контейнера (например, 'rentsoft_showcase')
     * @param  string $src         Полный URL страницы, которая будет открыта в iframe (значени атрибута src)
     * @param  string $width       Ширина iframe'а (по умолчанию 100%)
     * @return string              Код генерации элемента iframe
     */
    public static function getStretchingIframe($id, $src, $width)
    {
        if (!$width) $width = '100%';
        $src = addcslashes($src, "\\'\"\n\r");
        return trim("
            <div id='{$id}_div'></div>
            <script type='text/javascript'>
                (function () {
                    var ifr = document.createElement('iframe');
                    ifr.src = '{$src}' + '?ag_ref=' + escape(top.location.href.replace(/#.*/, ''));
                    ifr.id = '{$id}_iframe';
                    ifr.frameBorder = ifr.scrolling = 'no';
                    ifr.style.padding = ifr.style.margin = '0';
                    ifr.style.overflow = 'hidden';
                    ifr.style.height = '500px'; // initial height
                    ifr.style.width = '{$width}';
                    document.getElementById('{$id}_div').appendChild(ifr).onload = function() {
                        var prevH = null;
                        setInterval(function() {
                            var h = null;
                            if (!h) if (location.hash.match(/^#h(\d+)/)) h = RegExp.$1;
                            if (!h) {
                                // Delete frame if it has changed its window.name.
                                try { delete(top.frames['h' + prevH]) } catch (e) {}
                                if (top.frames['h' + prevH]) {
                                    // Cross-domain magic: if the marker still exists, it is real.
                                    h = prevH;
                                } else for (var i = 0; i < 10000; i += 30) if (top.frames['h' + i]) {
                                    // Previous marker disappeared: try to find a new one.
                                    h = i;
                                    break;
                                }
                            }
                            if (h && h != prevH) {
                                ifr.style.height = parseInt(h) + 'px';
                                ifr.style.width = '{$width}';
                                prevH = h;
                            }
                        }, 100);
                    };
                })();
            </script>
        ");
    }

    /**
     * Вызывается на стороне RentSoft.
     * Исходный код приведён для примера
     *
     * Возвращает javascript-код, реализующий поддержку автоматической подстройки высоты,
     * если страница открыта через iframe
     *
     * @return string
     */
    public static function getStretchingIframeRSSide()
    {
        return trim("
            <script type='text/javascript'>
                (function() {
                    if (location.href.match(/[?&amp;]in_frame=1/)) return; // only for frameset cabinet
                    var ie = 0 /*@cc_on + @_jscript_version @*/;
                    var dt = 100;
                    var prevH = 0;
                    var timeout = null;
                    // reading ag_ref from query string
                    var agRef = (function () {
                        var query = window.location.search.substring(1);
                        var vars = query.split(/&|\\?/);
                        for (var i = 0; i < vars.length; i++) {
                            var pair = vars[i].split('=');
                            if (pair[0] == 'ag_ref') {
                                return decodeURIComponent(pair[1]);
                            }
                        }
                        return null;
                    })();

                    if (!agRef) {
                        // reading agRef from cookie
                        agRef = (function () {
                            var nameEQ = 'ag_ref' + '=';
                            var ca = document.cookie.split(';');
                            for(var i = 0; i < ca.length; i++) {
                                var c = ca[i];
                                while (c.charAt(0)==' ') c = c.substring(1,c.length);
                                if (c.indexOf(nameEQ) == 0) {
                                    return decodeURIComponent(c.substring(nameEQ.length,c.length));
                                }
                            }
                            return null;
                        })();
                    } else {
                        // we know agRef. Store it in cookie
                        document.cookie = 'ag_ref=' + escape(agRef) + '; path=/';
                    }

                    if (!agRef) {
                        // we doesn't have an agRef.
                        // do nothing (for IE we MUST have agRef in any case)
                        return;
                    }
                    var callback = function(first) {
                    	var bp = document.getElementById('bottom_point');
                    	if (bp) {
	                        var h = Math.ceil(bp.offsetTop / 30) * 30;
	                        if (prevH < h) {
	                            var innerCallback = function() {
	                                // We use setTimeout to reduce continuous smooth resize
	                                // (e.g. jQuery effects) into one parent window signal.
	                                if (!ie) window.name = 'h' + h;
	                                else top.location.replace(agRef + '#h' + h);
	                            }
	                            if (timeout) clearTimeout(timeout);
	                            if (first) innerCallback();
	                            else timeout = setTimeout(innerCallback, dt);
	                            prevH = h;
	                        }
	                    }
                        setTimeout(callback, dt / 2);
                    }
                    callback(1); // it is IMPORTANT to call it synchronously
                })();
            </script>
        ");
    }
    
    /**
     * Вызывается на стороне Агента.
     *
     * Реализует mutex, препятствующий осуществлению параллельных списаний по одному пользователю.
     * Файлы mutex'ов создаются в session_save_path, поэтому удаление устаревших файлов
     * будет осуществляться автоматически
     *
     * @param string $id
     * @return resource $mutex
     */
    public function aquireMutex($id)
    {
        $orig = $id;
        if (!ctype_alnum($id)) $id = md5($id);
        $fname = session_save_path() . "/sess_rs_{$id}.lck";
        $f = @fopen($fname, "a+") or die("Cannot open $fname: " . print_r(error_get_last(), 1));
        flock($f, LOCK_EX);
        return $f;
    }

    /**
     * Вызывается на стороне Агента
     *
     * Отпускает mutex.
     * Метод необходим для версий PHP 5.3.x - более ранние версии
     * отпускали mutex автоматически при уничтожении переменной с ресурсом
     *
     * @param resource $mutex
     */
    public function releaseMutex($mutex)
    {
        flock($mutex, LOCK_UN);
        fclose($mutex);
    }
}
