import importlib

from cms.include import config, models


# Раз ориентир по архитектуре на джангу с глобальной конфигурацией,
#   то используем глобальный реестр источников данных: ds_name -> data_driver
ds_registry = {}  # type: dict[str, BaseDataDriver]


def get_dd(ds_name):
    """Вернём драйвер по имени источника данных"""
    return ds_registry.get(ds_name)


def init_ds():
    """Инициализация всех источников данных из конфига"""
    for ds_name, ds_config in config.data_sources.items():
        if ds_name not in ds_registry:
            # self.logger.info('Add Data Provider %s, config: %s', dp_name, dp_conf)
            print(f'APP: Add Data Source {ds_name}, config: {ds_config}')
            ds_registry[ds_name] = new_dd(ds_config, name=ds_name)


def init_db():
    """Инициализация, миграция схемы БД."""
    for ds_name, ds_dd in ds_registry.items():
        print(f'Init DB {ds_name}')
        try:
            ds_dd.init_db()
        except NotImplementedError:
            print(f'    не реализовано.')


# DataDriver factory
def new_dd(ds_config, name=None):
    dd_mod_class = ds_config.get('class')
    if not dd_mod_class:
        dd_type = ds_config.get('type')
        if dd_type == 'sqlalchemy':
            dd_mod_class = 'cms.data_providers.sqlalchemy:SQLAlchemyDataDriver'
    if not dd_mod_class:
        raise ValueError(f'You must specify valid "type" or "class" in datasource configuration', name, ds_config)
    dd_module_name, dd_class_name = dd_mod_class.split(':', maxsplit=1)
    dp_module = importlib.import_module(dd_module_name)
    dp_class = getattr(dp_module, dd_class_name)
    return dp_class(ds_config=ds_config, name=name)


# flask app event
def before_request():
    for dd in ds_registry.values():
        dd.before_request()


# flask app event
def before_first_request():
    for dd in ds_registry.values():
        dd.before_first_request()


# flask app event
def teardown_request(exception=None):
    for dd in ds_registry.values():
        dd.teardown_request(exception=exception)


def make_models():
    """make_models hook"""
    for dd in ds_registry.values():
        dd.make_models(models)


class BaseDataDriver:
    """"""
    def __init__(self, *_args, ds_config=None, name=None, **_kwargs):
        self.config = ds_config
        self.name = name

    def get(self, model, *args, **kwargs):
        pass

    def create(self, instance, *args, **kwargs):
        pass

    def delete(self, instance, *args, **kwargs):
        pass

    def save(self, instance, *args, **kwargs):
        pass

    def list(self, model, *arg, **kwargs):
        pass

    def before_request(self):
        pass

    def before_first_request(self):
        pass

    def teardown_request(self, exception=None):
        pass

    def dp_model(self, cms_model):
        pass

    def cms_model(self, sa_model):
        pass

    def is_instance(self, dp_instance):
        pass

    def make_models(self, models_):
        pass

    def query(self, *args, **kwargs):
        """Вернём кастомный запрос"""

    def init_db(self):
        """Создание/миграция схемы БД."""
        raise NotImplementedError


class Mapper:
    """Отображение CMS объектов в разный формат."""
    @classmethod
    def object_to_cms(cls, obj, cms_entity=None, model=None, model_name=None):
        """
        отображение obj -> cms_entity
        если cms_entity не указан, то создадим новый на основе модели cms_model или её имени model_name
        возвращаем ссылку на cms_entity
        """
        if not cms_entity:
            if not model:
                model = getattr(models, model_name)
            cms_entity = model()
        if not model:
            model = type(cms_entity)
        for f in model.Meta.fields:
            setattr(cms_entity, f, getattr(obj, f, None))
        return cms_entity

    @classmethod
    def cms_to_object(cls, cms_entity, obj):
        """
        отображение obj -> cms_entity
        если cms_entity не указан, то создадим новый на основе модели cms_model или её имени model_name
        возвращаем ссылку на cms_entity
        """
        model = type(cms_entity)
        for f in model.Meta.fields:
            setattr(obj, f, getattr(cms_entity, f))
        return obj


class DataProvider:
    """Объект, связывающий модель с источником данных"""
    def __init__(self, data_source=None, model=None):
        self.model = model
        self.data_source = data_source  # type: str
        self._dd = None  # type: BaseDataDriver

    @property
    def data_driver(self):
        if not self._dd:
            self._dd = get_dd(self.data_source)
        return self._dd

    def get(self, *args, **kwargs):
        return self.data_driver.get(self.model, *args, **kwargs)

    def create(self, *args, **kwargs):
        return self.data_driver.create(*args, **kwargs)

    def save(self, *args, **kwargs):
        return self.data_driver.save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        """
        Logical delete. Реализация в провайдере или в модели?
        self.deleted = 1
        self.save()
        """
        return self.data_driver.delete(*args, **kwargs)

    def list(self, *args, **kwargs):
        return self.data_driver.list(self.model, *args, **kwargs)

    def query(self, *args, **kwargs):
        return QueryProxy(self.data_driver.query(*args, **kwargs), dd=self.data_driver)


class QueryProxy:
    """
    Оборачиваем кастомный запрос и добавляем возможность извлекать данные с автомаппингом.
    TODO: Проверить слайсы query[a:b]
    """
    def __init__(self, query, dd=None):
        self._query = query
        self._dd = dd

    def __getattr__(self, item):
        """Обёртка для методов, которые возвращают новый запрос на основании текущего."""
        def method_proxy(*args, **kwargs):
            new_query = getattr(self._query, item)(*args, **kwargs)
            return QueryProxy(new_query, dd=self._dd)
        return method_proxy

    def _map_value(self, value):
        """Автоконвертация объектов в CMS"""
        if self._dd and self._dd.is_instance(value):
            value = Mapper.object_to_cms(value, model=self._dd.cms_model(type(value)))  # fixme теряем datasource
        return value

    def _map_row(self, row):
        if isinstance(row, tuple):
            # теряем имена, если row их поддерживает
            return tuple(self._map_value(col) for col in row)
        else:
            return self._map_value(row)

    def all(self):
        """Метод извлечения всех данных из текущего запрос с маппингом."""
        # for row in self._query:
        #     yield self._map_row(row)
        return [self._map_row(row) for row in self._query]

    def first(self):
        """Метод извлечения первой строки с маппингом."""
        row = self._query.first()
        return self._map_row(row)

    @property
    def dd_query(self):
        return self._query

    def __iter__(self):
        """Чтение строк из запроса"""
        return iter(self.all())
