import inspect
from collections import OrderedDict

import sqlalchemy.ext.declarative
import sqlalchemy.orm

from cms import fields
from cms.include import models
from ..models.base_model import BaseModelMeta, DEFAULT_DATASOURCE
from .base import BaseDataDriver, Mapper


class SQLAlchemyDataDriver(BaseDataDriver):
    """
    dataproviders['default'] = {
        'type': 'sqlalchemy',
        'sqlachemy.url': 'postgresql://user:password@host/db',
        'sqlachemy.echo': True,  # Достаточно для отладки запросов в develop mode
    }

    # Для отладки в продакшене логгеры:
    sqlalchemy.engine - controls SQL echoing. set to
        logging.INFO for SQL query output, logging.
        DEBUG for query + result set output.
        These settings are equivalent to echo=True and echo="debug" on create_engine.echo, respectively.
    sqlalchemy.pool - controls connection pool logging.
         set to logging.INFO to log connection invalidation and recycle events;
         set to logging.DEBUG to additionally log all pool checkins and checkouts.
         These settings are equivalent to pool_echo=True and pool_echo="debug" on create_engine.echo_pool, respectively.
    sqlalchemy.dialects - controls custom logging for SQL dialects, to the extend that logging
         is used within specific dialects, which is generally minimal.
    sqlalchemy.orm - controls logging of various ORM functions to the extent that logging
         is used within the ORM, which is generally minimal.
         Set to logging.INFO to log some top-level information on mapper configurations.

    Example:
        logging.basicConfig()
        logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)
    """
    # Все модели проекта в глобальном реестре
    models_registry = {}
    db_meta = sqlalchemy.MetaData()
    sa_base_model = sqlalchemy.ext.declarative.declarative_base(class_registry=models_registry, metadata=db_meta)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.engine = sqlalchemy.engine_from_config(
            self.config,
            **{
                k.replace('sqlalchemy.', '', 1): v
                for k, v in kwargs.items() if k.startswith('sqlalchemy.')})
        self.Session = sqlalchemy.orm.scoped_session(sqlalchemy.orm.sessionmaker(bind=self.engine))

    @classmethod
    def __get_indexes(cls, cms_model):
        indexes = []
        # Пока только одноколоночные, мультиколоночные будем брать из меты
        for cms_field_name, cms_field in cms_model.Meta.fields.items():
            if getattr(cms_field, 'unique', None):
                indexes.append(('unique', (cms_field_name, )))
            elif getattr(cms_field, 'index', None):
                indexes.append(('index', (cms_field_name, )))
        # todo: мультиколоночные индексы cms_model.Meta.indexes -> indexes
        return indexes

    @classmethod
    def __make_sa_column(cls, _cms_model, cms_field, _db_field_name):
        """Создадим SA поле для модели"""
        # TODO timestamp, datetime, numeric
        if isinstance(cms_field, fields.IntegerField):
            sa_type = sqlalchemy.Integer
        elif isinstance(cms_field, fields.StringField):
            sa_type = sqlalchemy.String
        else:
            sa_type = sqlalchemy.String
        column_kwargs = {}
        return sqlalchemy.Column(sa_type, **column_kwargs)

    @classmethod
    def dp_model_cls(cls, cms_model):
        """
        Вернём SA модель для CMS модели.
        TODO: db_name
        """
        model_name = cms_model.__name__
        sa_model = cls.models_registry.get(model_name)
        if not sa_model:
            # модель отсутствует, создадим её
            print(f'Make SaModel for {model_name}')  # fixme
            tablename = getattr(getattr(cms_model, 'Meta', None), 'tablename', None)
            if not tablename:
                tablename = model_name

            sa_columns = []
            __table_args__ = []
            __table_kwargs__ = {}
            # Добавим поля
            pk_columns = cms_model.Meta.primary_key
            # print(f'Pk columns {pk_columns}')
            for cms_field_name, cms_field in cms_model.Meta.fields.items():
                sa_field_name = cms_field_name
                sa_column = cls.__make_sa_column(cms_model, cms_field, sa_field_name)
                # Доп свойства колонок
                if cms_field_name in pk_columns:
                    sa_column.primary_key = True
                if cms_field.nullable is False:
                    sa_column.nullable = False
                if cms_field.default is not None:
                    sa_column.default = cms_field.default
                sa_column.name = sa_field_name
                print(f'Add field {model_name}.{sa_field_name}')  # fixme
                sa_columns.append(sa_column)
            # Добавим индексы
            for index_type, index_tuple in cls.__get_indexes(cms_model):
                print(f'Add index {index_tuple} {index_type}')
                if index_type == 'unique':
                    # __table_args__.append(sqlalchemy.UniqueConstraint(None, *index_tuple))
                    __table_args__.append(sqlalchemy.Index(None, *index_tuple, unique=True))
                else:  # must be index
                    __table_args__.append(sqlalchemy.Index(None, *index_tuple))
            # Создадим модель
            __table_args__.append(__table_kwargs__)  # __table_kwargs__ must be last element in args tuple
            sa_model_kwargs = {'__tablename__': tablename, '__table_args__': tuple(__table_args__)}
            sa_model_kwargs.update({c.name: c for c in sa_columns})
            sa_model = type(model_name, (cls.sa_base_model,), sa_model_kwargs)
            cls.models_registry[model_name] = sa_model
        return sa_model

    def dp_model(self, cms_model):
        return self.dp_model_cls(cms_model)

    def cms_model(self, sa_model):
        """
        Вернём CMS модель для SA модели.
        """
        return getattr(models, sa_model.__name__)

    def is_instance(self, dp_instance):
        return isinstance(dp_instance, SQLAlchemyDataDriver.sa_base_model)

    # ------------- Data Provider API ---------------------
    def list(self, model, *arg, **kwargs):
        """Грубо селектим всю таблицу, фильтры:
          - в kwargs: поле -> значение
        """
        s = self.Session()
        sa_model = self.dp_model(model)
        q = s.query(sa_model)
        for k, v in kwargs.items():
            q = q.filter(getattr(sa_model, k) == v)
        return [Mapper.object_to_cms(instance, model_name=sa_model.__name__) for instance in q]

    def get(self, model, *args, **kwargs):
        """
        Выбрать один объект по фильтру:
          - в args - первичный ключ(опционально)
          - в kwargs - другие поля: поле -> значение"""
        s = self.Session()
        sa_model = self.dp_model(model)

        if len(args) > 0:
            pkey = model.Meta.primary_key
            for i in range(len(args)):
                kwargs[pkey[i]] = args[i]

        sa_instance = s.query(sa_model).filter(
            *[getattr(sa_model, k) == v for k, v in kwargs.items() if k in model.Meta.fields]
        ).first()

        if sa_instance:
            return Mapper.object_to_cms(sa_instance, model_name=sa_model.__name__)

    def create(self, instance, *args, **kwargs):
        s = self.Session()
        sa_model = self.dp_model(instance.__class__)
        sa_instance = sa_model()
        Mapper.cms_to_object(instance, sa_instance)
        s.add(sa_instance)
        s.flush(objects=[sa_instance])  # force insert
        # обратно копируем значения, если БД их сгенерила-поменяла.
        Mapper.object_to_cms(sa_instance, cms_entity=instance)

    def delete(self, instance, *args, **kwargs):
        """TODO логическое удаление"""
        s = self.Session()
        sa_model = self.dp_model(instance.__class__)
        pkey = type(instance).Meta.primary_key
        # 'get' делает запрос только, объекта нет у сессии
        sa_instance = s.query(
            sa_model).get(tuple((getattr(instance, f) for f in pkey)))
        s.delete(sa_instance)

    def save(self, instance, *args, **kwargs):
        s = self.Session()
        sa_model = self.dp_model(type(instance))
        pkey = type(instance).Meta.primary_key
        sa_instance = s.query(
            sa_model).get(tuple((getattr(instance, f) for f in pkey)))
        Mapper.cms_to_object(instance, sa_instance)
        # Надо ли форсить update?
        s.flush(objects=[sa_instance])  # force update
        # обратно копируем значения, если БД их сгенерила-поменяла.
        Mapper.object_to_cms(sa_instance, cms_entity=instance)

    def before_request(self):
        super().before_request()

    def before_first_request(self):
        """TODO: Create schema"""
        super().before_first_request()

    def teardown_request(self, exception=None):
        if exception:
            self.Session().rollback()
        else:
            self.Session().commit()
        self.Session.remove()
        super().teardown_request(exception=exception)

    def query(self, *args, **kwargs):
        return self.Session().query(*args, **kwargs)

    @classmethod
    def make_models(cls, models_):
        super().make_models(models_)
        # Не будем создавать модели статически в файле

    def init_db(self):
        """Создадим схему"""
        # Нужно инициализировать все модели, чтобы заполнить мету
        for model_name in dir(models):
            cms_model = getattr(models, model_name)
            # не модель
            if not isinstance(cms_model, BaseModelMeta) or cms_model.Meta.abstract:
                continue
            # print(f'    Check model meta: {cms_model.Meta}')
            # Модель явно с нами не работает
            if cms_model.Meta.data_sources and self.name not in cms_model.Meta.data_sources:
                continue
            # мы не дефолт источник
            if not cms_model.Meta.data_sources and self.name != DEFAULT_DATASOURCE:
                continue
            print(f'    Add model meta: {cms_model.Meta}')
            self.dp_model(cms_model)  # create model and meta
        # Создадим таблицы(пока без миграции)
        #   create_all - создаёт таблицу, если её нет. Колонки и всё остальное не смотрит.
        self.db_meta.create_all(self.engine)
        # TODO: сделать миграцию через рефлекшен


class SAModelAccessor:
    def __getattr__(self, item):
        return SQLAlchemyDataDriver.dp_model_cls(getattr(models, item))
