from cms.include import *
import string
import random
import time


class FieldValidateError(Exception):
    def __init__(self, error, field, value):
        self.error = error
        self.field = field
        self.value = value

    def __str__(self):
        return 'Невозможно присвоить значение {} для поля "{}" типа {}. {}'\
            .format(self.value, self.field.caption, self.field.__class__.__name__, self.error)


class BaseField(object):
    def __init__(self,
                 caption,
                 placeholder=None,
                 comment=None,
                 nullable=None,
                 primary_key=None,
                 unique=None,
                 index=None,
                 db_name=None,
                 db_alias=None,
                 default=None):
        """
            caption - наименование поля (русский текст)
            placeholder - placeholder для html виджетов
            comment - поясняющий текст, в дополнение к caption
            db_name - имя поля в БД (По-дефолту - имя поля) для полей с несколькими источниками - можно передавать словарь
            db_alias - алиас источника/БД, когда модель является собранной из нескольких источников
            index - необхидимо создать индекс по полю
            primary_key - поле входит в первичный ключ
            unique - необхидимо создать уникальный индекс по полю
        """
        self.caption = caption
        self.placeholder = placeholder
        self.comment = comment
        self.name = None  # Имя из модели
        self.base_model = None  # Имя модели от которой унаследовали
        self.model = None  # Имя модели
        self.db_name = db_name
        self.db_alias = db_alias
        # TODO: вернуть проверку
        self.nullable = True
        self.default = default
        self.primary_key = primary_key
        self.unique = unique
        self.index = index
        # system

    def __get__(self, obj, objtype):
        if obj is None:
            return self
        if self.name in obj.__dict__:
            return obj.__dict__[self.name]
        else:
            if self.default is None and not self.nullable:
                raise Exception('Укажите для поля {} default, либо выставите ему nullable=True'.format(self.name))
            # Поддержка lambda в default
            if callable(self.default):
                return self.default()
            return self.default

    def __set__(self, obj, val):
        self.validate(val)
        # У модели выставляем флаг changed
        obj._changed = True
        obj.__dict__[self.name] = val

    def validate(self, val):
        if val is None:
            if self.nullable:
                return
                self._raise_validation_error(val, '{} не может быть None!'.format(self.caption))

    def _raise_validation_error(self, val, msg=''):
        raise FieldValidateError(msg, self, val)

    def __set_name__(self, model, field_name):
        """Вызывает фабрика модели, чтобы передать имя и хозяина.
         Начиная с 3.6, для этого существует метод дескриптора __set_name__(self, owner, name)"""
        if not self.name:
            self.name = field_name
        if not self.db_name:
            self.db_name = self.name
        if not self.base_model:
            self.base_model = model
        self.model = model


class IntegerField(BaseField):
    def validate(self, val):
        super().validate(val)
        if val is None:
            return
        if not isinstance(val, int):
            self._raise_validation_error(val, 'Должен быть int')


class StringField(BaseField):
    def __init__(self, *args, maxlen=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.maxlen = maxlen

    def validate(self, val):
        super().validate(val)
        if val is None:
            return
        if isinstance(val, str):
            if self.maxlen and len(val) > self.maxlen:
                self._raise_validation_error(val, f'Длина строки превышает лимит {self.maxlen}')
            return
        self._raise_validation_error(val, 'Должен быть str')


class String32Field(StringField):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, maxlen=32, **kwargs)


class String256Field(StringField):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, maxlen=256, **kwargs)


class String4096Field(StringField):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, maxlen=4096, **kwargs)


class PhoneNumberField(String32Field):
    def validate(self, val):
        super().validate(val)
        if val and not re.match(r'^\+?[1-9]\d{1,14}$', val):  # see E.164
            self._raise_validation_error(val, 'Неправильный формат телефона')


class EmailField(String256Field):
    def validate(self, val):
        super().validate(val)
        if val and not re.match(r'.+@.+\..+', val):
            self._raise_validation_error(val, 'Введите правильный email адрес')


class IpField(String32Field):
    def validate(self, val):
        super().validate(val)
        if val and not re.match(r'^([0-9]{1,3}\.){3}[0-9]{1,3}(/([0-9]|[1-2][0-9]|3[0-2]))?$', val):
            # Matches ipv4 with and without subnet, e.g. 192.168.0.1 and 192.168.0.1/32
            # SEE: http://www.regexpal.com/93987
            self._raise_validation_error(val)


class GeneratedPasswordField(String32Field):
    def __init__(self, caption, password_len, *args, **kwargs):
        super().__init__(caption, *args, **kwargs)
        self.password_len = int(password_len)

    @property
    def default_value(self):
        alphabet = string.ascii_letters + string.digits
        return ''.join(random.choice(alphabet) for _ in range(self.password_len))


class TimestampField(BaseField):
    def validate(self, val):
        super().validate(val)
        if val is None:
            return
        # TODO: change to datetime.now()
        if not isinstance(val, int):
            self._raise_validation_error(val, 'Должен быть int')

    @property
    def default_value(self):
        return int(time.time())


class BooleanField(BaseField):
    def validate(self, val):
        super().validate(val)
        if val is None:
            return
        if not isinstance(val, bool):
            self._raise_validation_error(val, 'Должен быть bool')


class ListField(BaseField):
    def validate(self, val):
        super().validate(val)
        if val is None:
            return
        if not isinstance(val, list):
            self._raise_validation_error(val, 'Должен быть list')


class DecimalField(BaseField):
    '''
    TODO
    Планирую реализовать возможность при установке значения, если передали строку - кормить ее в конструктор Decimal.
    Чтобы суммы и деньги могли прописывать строкой:
    product.price = D('315.50')
    vs
    product.price = '315.50'
    Понадобилось после передачи Decimal через json (который его не умеет, но умеет его переделывать в строку)

    Возможно, возникнет путаница, когда полю присваивают строку, а в значении оказывается Decimal. Будет путать 100%
    От этого поля можно будет отнаследовать поля MoneyField и т.п.
    '''
    pass
