from cms.include_cms import *
from collections import OrderedDict
from cms.fields.base_fields import BaseField, IntegerField, String256Field


DEFAULT_DATASOURCE = 'default'


class ModelMeta:
    """
    Метаданные модели.
    """
    def __init__(self):
        self.model = None  # type: BaseModel
        self.fields = None  # type: OrderedDict
        self.tablename = None  # type: str
        self.data_sources = None  # type: tuple  # 'default', 'rest'
        self.abstract = None  # type: bool  # Признак Абстрактной модели. Для наследования.
        # todo: мультиколоночные индексы cms_model.Meta.indexes -> indexes

    @property
    def primary_key(self):
        explicit_pkey = tuple((f_name for f_name, f in self.fields.items() if f.primary_key))
        if explicit_pkey:
            return explicit_pkey
        if 'id' in self.fields:
            return 'id',

    def __repr__(self):
        return f'ModelMeta(Model={self.model.__module__}.{self.model.__qualname__}, tablename={self.tablename},' \
            f' abstract={self.abstract}, data_sources={self.data_sources})'


class BaseModelMeta(type):
    """
    Model class builder
    """
    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # for auto completion and type hints
        cls.__meta__ = getattr(cls, '__meta__')  # type: ModelMeta
        # датапровайдеры модели: 'datasource' -> dataprovider('datasource', ModelCls)
        cls.__dp__ = {}  # type: dict[str, DataProvider]

    @property
    def Meta(cls):
        """метаданные модели"""
        return cls.__meta__  # type: ModelMeta

    @property
    def dp(cls):
        """Доступ к датапровайдеру модели"""
        from cms.data_providers.base import DataProvider  # lazy load

        data_source = cls.__meta__.data_sources[0] if cls.__meta__.data_sources else DEFAULT_DATASOURCE
        dp_ = cls.__dp__.get(data_source)  # type: DataProvider
        if not dp_:
            dp_ = DataProvider(data_source=data_source, model=cls)
            cls.__dp__[data_source] = dp_
        return dp_

    @staticmethod
    def __new__(mcs, name, bases, namespace, **kwargs):
        """Фабрика моделей
        Неплохое место, чтобы датасхему формировать.
        Хотя дата схему, можно сделать в любой момент инспектом modules.*
        """
        meta = ModelMeta()
        meta.fields = OrderedDict()
        # Соберём конфигурацию базовых классов
        for base_cls in bases:
            # Соберём реестр полей.
            base_meta = getattr(base_cls, 'Meta', None)  # type: ModelMeta
            if base_meta:
                if base_meta.tablename:
                    meta.tablename = base_meta.tablename
                if base_meta.fields:
                    for field_name, field in base_meta.fields.items():
                        meta.fields[field_name] = field
                if base_meta.data_sources:
                    meta.data_sources = tuple(base_meta.data_sources)
        # Собственная конфигурация
        for attr_name, attr_value in namespace.items():
            # Добавим собственные поля в реестр
            if isinstance(attr_value, BaseField):
                # attr_value.set_model_name(name, attr_name)  # descriptor.__set_name__(self, owner, name) - since 3.6
                meta.fields[attr_name] = attr_value
            if attr_name == 'Meta':
                cfg_meta = attr_value
                meta.tablename = getattr(cfg_meta, 'tablename', meta.tablename)  # признак может наследоваться
                meta.abstract = getattr(cfg_meta, 'abstract', not meta.tablename)  # признак не наследуется fixme
                meta.data_sources = getattr(cfg_meta, 'data_sources', meta.data_sources)  # признак может наследоваться
        # конфигурационный класс не нужен, доступ к реальной мете через свойство.
        if 'Meta' in namespace:
            del namespace['Meta']
        # fixme abstract
        if not meta.tablename:
            meta.abstract = True
        # Неявное имя таблицы(возможно лучше всегда явное)
        if not meta.tablename and not meta.abstract:
            meta.tablename = name
        namespace['__meta__'] = meta  # метаданные модели
        new_cls = super().__new__(mcs, name, bases, namespace, **kwargs)
        meta.model = new_cls
        return new_cls


class BaseModel(metaclass=BaseModelMeta):
    class Meta:
        abstract = True

    def __init__(self, *_args, **kwargs):
        """
        Инициализация вида Object(name='John', address='BigStreet')
        TODO: datasource?
        """
        for attr, value in kwargs.items():
            setattr(self, attr, value)

    @property
    def values_dict(self):
        """Возвращает значения всех полей в виде dict"""
        return OrderedDict(((field_name, getattr(self, field_name)) for field_name in type(self).Meta.fields))

    @classmethod
    def get(cls, *args, **kwargs):
        return cls.dp.get(*args, **kwargs)

    def create(self, *args, **kwargs):
        return type(self).dp.create(self, *args, **kwargs)

    def save(self, *args, **kwargs):
        return type(self).dp.save(self, *args, **kwargs)

    def delete(self, *args, **kwargs):
        """
        Logical delete. Реализация в провайдере или в модели?
        self.deleted = 1
        self.save()
        """
        return type(self).dp.delete(self, *args, **kwargs)

    @classmethod
    def list(cls, *args, **kwargs):
        return cls.dp.list(cls, *args, **kwargs)

    def __repr__(self):
        fields = ', '.join(f'{f}={getattr(self, f)}' for f in type(self).Meta.fields)
        return f'models.{type(self).__name__}({fields})'


class CmsModel(BaseModel):
    """Базовый класс, содержит стандартные поля для типовых объектов."""
    class Meta:
        abstract = True

    id = IntegerField('id объекта', primary_key=True)
    name = String256Field('Имя объекта', index=True)
