"""
Cross Site Request Forgery Middleware.

This module provides a middleware that implements protection
against request forgeries from other sites.
"""
import string
import random
import hmac
from flask import session, request, abort

CSRF_SECRET_LENGTH = 32
CSRF_ALLOWED_CHARS = string.ascii_letters + string.digits


def _get_new_csrf_string():
    return ''.join(random.choice(CSRF_ALLOWED_CHARS) for i in range(CSRF_SECRET_LENGTH))


def _salt_cipher_secret(secret):
    salt = _get_new_csrf_string()
    chars = CSRF_ALLOWED_CHARS
    pairs = zip((chars.index(x) for x in secret), (chars.index(x) for x in salt))
    cipher = ''.join(chars[(x + y) % len(chars)] for x, y in pairs)
    return salt + cipher


def _unsalt_cipher_token(token):
    salt = token[:CSRF_SECRET_LENGTH]
    token = token[CSRF_SECRET_LENGTH:]
    chars = CSRF_ALLOWED_CHARS
    pairs = zip((chars.index(x) for x in token), (chars.index(x) for x in salt))
    secret = ''.join(chars[x - y] for x, y in pairs)  # Note negative values are ok
    return secret


def _get_new_csrf_token():
    return _salt_cipher_secret(_get_new_csrf_string())


def get_csrf_token():
    if "csrf_token" not in session:
        csrf_secret = _get_new_csrf_string()
        session["csrf_token"] = _salt_cipher_secret(csrf_secret)
    else:
        csrf_secret = _unsalt_cipher_token(session["csrf_token"])
    return _salt_cipher_secret(csrf_secret)


def rotate_csrf_token():
    session["csrf_token"] = _get_new_csrf_token()


def _compare_salted_tokens(request_csrf_token, csrf_token):
    # Assume both arguments are sanitized -- that is, strings of
    # length CSRF_TOKEN_LENGTH, all CSRF_ALLOWED_CHARS.
    return hmac.compare_digest(
        _unsalt_cipher_token(request_csrf_token),
        _unsalt_cipher_token(csrf_token),
    )


def csrf_token_required(f):
    def wrapper(*args, **kwargs):
        session_token = session.get("csrf_token", None)
        form_token = request.form.get("csrf_token", None)
        if not session_token:
            rotate_csrf_token()
            abort(400, "No csrf token is session")
        if not form_token:
            abort(400, "No csrf token in form")
        if not _compare_salted_tokens(session_token, form_token):
            abort(400, "csrf token is not valid")
        return f(*args, **kwargs)

    return wrapper
